package org.vraptor.component;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.vraptor.annotations.Component;
import org.vraptor.annotations.Destroy;
import org.vraptor.annotations.In;
import org.vraptor.annotations.Parameter;
import org.vraptor.annotations.Read;
import org.vraptor.interceptor.InterceptorType;
import org.vraptor.introspector.FieldParameter;
import org.vraptor.introspector.FieldReadParameter;
import org.vraptor.introspector.ReadParameter;
import org.vraptor.reflection.ReflectionUtil;
import org.vraptor.reflection.StringUtil;
import org.vraptor.scope.ScopeType;

/**
 * A component class factory.
 * 
 * @author Guilherme Silveira
 */
public class ComponentTypeFactory {

	static final String VALIDATE_METHOD_INITIALS = "validate";

	private static final Logger LOG = Logger.getLogger(ComponentTypeFactory.class);

	private final LogicMethodFactory factory = new LogicMethodFactory();

	public static final String[] COMPONENT_TERMINATIONS = { "Controller", "Logic", "Command", "Action", "Component"};

	/**
	 * Creates a component class dealer
	 * 
	 * @param type
	 *            the class to load
	 * @return
	 * @throws InvalidComponentException
	 */
	public ComponentType getComponentType(Class<?> type) throws InvalidComponentException {
		String componentName = getComponentName(type);

		if (!type.isAnnotationPresent(Component.class)) {
			LOG.warn("Deprecated: registering " + type.getName() + " component without @Component annotation.");
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("Component found: " + componentName + " from " + type.getName());
		}

		ScopeType scope = getScope(type);

		Clazz clazz = new Clazz(type);
		BeanConstructor constructor = clazz.findSingleConstructor();
		LOG.debug("Registered constructor: " + constructor);

		Map<String, DefaultLogicMethod> actions = factory.loadLogics(type);

		List<InterceptorType> interceptors = InterceptorType.getInterceptors(type);

		// read fields
		List<FieldAnnotation<In>> ins = ReflectionUtil.readAnnotations(type, In.class);

		// destroy method
		String destroyLogicName = getDestroyLogicName(type);

		LOG.debug("Component clazz " + type.getName() + " read");

		List<ReadParameter> reads = findParameters(type);

		return new DefaultComponentType(type, componentName, scope, constructor, actions, ins, interceptors,
				destroyLogicName, reads);

	}

	List<ReadParameter> findParameters(Class<?> type) {
		ArrayList<ReadParameter> reads = new ArrayList<ReadParameter>();
		for (FieldAnnotation<Read> info : ReflectionUtil.readAnnotations(type, Read.class)) {
			reads.add(new FieldReadParameter(info));
		}
		for (FieldAnnotation<Parameter> info : ReflectionUtil.readAnnotations(type, Parameter.class)) {
			reads.add(new FieldParameter(info));
		}
		return reads;
	}

	String getDestroyLogicName(Class<?> type) {
		String destroyLogicName = "destroy";
		for (Method m : type.getMethods()) {
			if (m.isAnnotationPresent(Destroy.class)) {
				destroyLogicName = m.getName();
			}
		}
		return destroyLogicName;
	}

	ScopeType getScope(Class<?> type) {
		return type.isAnnotationPresent(Component.class) ? type.getAnnotation(Component.class).scope()
				: ScopeType.LOGIC;
	}

	String getComponentName(Class<?> type) {
		String componentName;

		if (type.isAnnotationPresent(Component.class)) {
			Component ann = ((Component) type.getAnnotation(Component.class));
			if (!ann.value().equals("")) {
				componentName = ann.value();
			} else {
				String name = StringUtil.removeEnding(type.getSimpleName(), COMPONENT_TERMINATIONS);

				if (!name.equals(type.getSimpleName())) {
					componentName = name.toLowerCase();
				} else {
					componentName = type.getSimpleName();
				}
			}
		} else {
			componentName = type.getSimpleName();
		}

		return componentName;
	}

}
